%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Mise en oeuvre du filtre de Kalman pour l'estimation de la
% position d'un chariot a roues encodeuses avec capteur laser embarque 
%
% Souce: Cours GLO-4001/GLO-7021, Philippe Giguere, Universite de Laval, Canada (2010-2012)
% www2.ift.ulaval.ca/~pgiguere/cours/IntroRobotique/index.html
% (adapte par F. Morbidi, Jan. 2019)
%
% TD2, Exercice II - Surveillance Distribuee de Systemes Multi-agents 
% ---------------------------------------------------------------------

clear
close all
clc

% Variables decrivant le systeme dynamique
Sigma_odom = 0.01; 
Sigma_laser = 0.1; 
Q = Sigma_odom^2;  % Variance du bruit de modele (odometrie)
R = Sigma_laser^2; % Variance du bruit de mesure (laser)

xVrai = 0; % Initialisation de l'etat reel (position du chariot)
x = xVrai; % Initialisation de l'etat estime 
P = 0.1;   % Initialisation de la variance de l'erreur d'estimation 
    
% On traca les distributions des deux bruits (assumes gaussiens)

myFontSize = 13; % Pour changer la police dans les figures

%%% Affichage des erreurs de deplacement et de mesure  

Bins = -0.4:0.001:0.4;
PasDistrib = Sigma_odom*randn(1,100000);
figure(1),
subplot(2,1,1),
hist(PasDistrib,Bins)
xlim([-0.4 0.4]);
c = axis();
title('Distribution des erreurs sur le deplacement','FontSize',myFontSize);
xlabel('Erreur [m]','FontSize',myFontSize);
set(gca,'YTick',[]);
text(-0.3,4000,sprintf('sigma = %.2f m',Sigma_odom),'FontSize',myFontSize-1);
set(gca,'FontSize',myFontSize);

subplot(2,1,2),
LaserDistrib = Sigma_laser*randn(1,100000);
hist(LaserDistrib,Bins)
set(gca,'YTick',[]);
axis(c);
title('Distribution des erreurs sur les mesures laser','FontSize',myFontSize);
text(-0.3,4000,sprintf('sigma = %.2f m',Sigma_laser),'FontSize',myFontSize-1);
xlabel('Erreur [m]','FontSize',myFontSize);
set(gca,'FontSize',myFontSize);

V = 1;     % Vitesse du chariot [m/s] 
Dk = 0.1;  % Pas de discretisation [s]
Tmax = 50; % Nombre d'iterations du filtre de Kalman

display('Appuyez sur "Entree" pour continuer');
pause

figure(2)

%%% Mise en oeuvre du filtre de Kalman %%%

for iStep = 1:Tmax
    % Simulation du systeme dynamique (chariot + laser)
    xVrai = xVrai + V*Dk + Sigma_odom*randn;
    z  = xVrai + Sigma_laser*randn; % mesure laser
    
    % ==== Prediction du filtre de Kalman ====
    x  = x + V*Dk;
    P  = P + Q;
    
    % ==== Correction du filtre de Kalman ====
    K  = P/(P + R); % Gain de Kalman
    x  = x + K*(z-x);
    P  = (1-K)*P;
    
    % Stockage de l'etat reel et estime, de la variance P, etc. et des erreurs d'estimation/mesure
    Ax(iStep) = xVrai;
    AEx(iStep) = x;
    AP(iStep) = P;
    AK(iStep) = K;
    ALaser(iStep) = z;
    ErreurEx(iStep) = abs(x-xVrai);    % Module de l'erreur d'estimation de la position du chariot
    ErreurLaser(iStep) = abs(z-xVrai); % Module de l'erreur de mesure
    ATime(iStep) = iStep*Dk;
    
    subplot(2,1,1);
    plot(ATime,ALaser,'r*');
    xlabel('temps [s]','FontSize',myFontSize);
    ylabel('Mesure laser [m]','FontSize',myFontSize);
    set(gca,'FontSize',myFontSize);

    subplot(2,1,2);
    h(1) = plot(ATime,Ax,'k+','LineWidth',2);
    hold on;
    h(3) = plot(ATime,AEx,'go','MarkerFaceColor','g');
    h(2) = plot(ATime,ALaser,'r*');
    xlabel('temps [s]','FontSize',myFontSize);
    ylabel('Position [m]','FontSize',myFontSize);
    legend(h,{'Position reelle' 'Mesure laser' 'Position estimee (filtre de Kalman)'},'Location','NorthWest');
    set(gca,'FontSize',myFontSize);
    pause(0.5)
end

% Affichage de l'evolution temporelle du gain de Kalman K 
% et de la variance de l'erreur d'estimation P

figure(3)
subplot(2,1,1)
plot(ATime,AK,'k-','LineWidth',2)
xlabel('temps [s]','FontSize',myFontSize)
ylabel('Gain de Kalman K','FontSize',myFontSize)
title('Evolution temporelle de K (en haut) et de P (en bas)')
set(gca,'FontSize',myFontSize);
xlim([0 Tmax/10])

subplot(2,1,2)
plot(ATime,AP,'k-','LineWidth',2)
xlabel('temps [s]','FontSize',myFontSize)
ylabel('Variance P','FontSize',myFontSize)
set(gca,'FontSize',myFontSize)
xlim([0 Tmax/10])

disp('                                                             ')
disp('Moyenne de l''erreur d''estimation de la position du chariot:')
mean(ErreurEx)     
disp('Moyenne de l''erreur de mesure:')
mean(ErreurLaser) 
