%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Mise en oeuvre du filtre de Kalman etendu (EKF) pour l'estimation de l'etat 
% d'un drone (position, vitesse, acceleration), a partir des mesures de distance  
% de n balises 3D, s1,s2,...,sn, de position connue 
%
% TD2, Exercice IV - Surveillance Distribuee de Systemes Multi-agents 
% --------------------------------------------------------------------------------

clear
close all
clc

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Parametres du systeme dynamique %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

delta = 0.1; % Pas de discretization [s]

T = 50; % Nombre d'iterations de l'EKF 

%% Matrices du systeme dynamique

A = [eye(3), delta*eye(3), (delta^2/2)*eye(3); zeros(3), eye(3), delta*eye(3); zeros(3), zeros(3), eye(3)];

B = [(delta^2/4)*eye(3); (delta/2)*eye(3); eye(3)];

F = [(delta^2/2)*eye(3); delta*eye(3); zeros(3)];

%% Bruit de modele et de mesure 

sigma_w = 1e-3; 
Qp = sigma_w^2*eye(3); 

sigma_r = 0.2;

% ----------------------------------------------------------------
% ATTENTION: nous avons fait ici deux hypotheses simplificatrices
% ----------------------------------------------------------------
% 1) La variance du bruit de mesure est proportionnelle   
%    a la distance des balises. Pour plus de simplicite, nous   
%    avons considere la meme variance pour toutes les mesures de distance
% 2) Les capteurs de distance ont une portee infinite 
               
%% Positionnement aleatoire des balises dans l'espace 3D

%% On considere un cube de cote L qui contient toutes les balises 

L = 2;
xMin = 0; xMax = L;
yMin = 0; yMax = L;
zMin = 0; zMax = L;

xDelta = xMax-xMin; 
yDelta = yMax-yMin; 
zDelta = zMax-zMin; 

numbPoints = 100; % Nombre de balises

xx = xDelta*(rand(numbPoints,1)) + xMin;  
yy = yDelta*(rand(numbPoints,1)) + yMin;  
zz = zDelta*(rand(numbPoints,1)) + zMin;  

S = [xx, yy, zz]; % Les coordonnees 3D des balises sont stockees dans le colonnes de la matrice S

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Etat initial du drone: vecteur x0 = [p0; v0; a0] de taille 9 x 1 %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

x0 = zeros(9,1); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Initialisation de l'EKF  %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

xhat = x0 + 0.5*rand(9,1); % Etat estime initial
P = eye(9); % Matrice de covariance de l'erreur d'estimation initiale 

%% Generation du vecteur d'entree (variation des trois accelerations selon x,y et z)

for k = 1:T
    u(:,k) = [0.01; 0.02*sin(delta*k); 0.01]; 
end

%% Matrice de covariance du bruit de mesure

R = diag(sigma_r^2*ones(numbPoints,1)); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Mise en oeuvre de l'EKF %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

xtrue_store = []; % Matrice pour stocker les vecteurs d'etat vrais
xhat_store = [];  % Matrice pour stocker les vecteurs d'etat estimes
P_store = [];     % Matrice pour stocker les matrices de covariance
Time_store = [];  % Vecteur pour stocker les temps

xtrue = x0;

for k = 1:T
    %% Simulation du systeme dynamique 
    xtrue_new = A*xtrue + B*u(:,k) + F*sigma_w*randn(3,1);
     
    %% ==== Prediction de l'EKF ==== 
    xhat = A*xhat + B*u(:,k);
    P = A*P*A' + F*Qp*F';
    
    %% ==== Correction de l'EKF ==== 
    z = []; 
    H = []; 
    h = [];
    for i = 1:numbPoints % Toutes les measures sont rassemblees
        z = [z; norm(xtrue_new(1:3)' - S(i,:)) + sigma_r*randn];
        H = [H; (xhat(1:3)' - S(i,:))/norm(xhat(1:3)' - S(i,:)), zeros(1,6)];
        h = [h; norm(xhat(1:3)' - S(i,:))];
    end
         
    K = P*H'*inv(H*P*H' + R); % Gain de Kalman
    xhat = xhat + K*(z - h);  % Mise a jour de l'etat estime
    P = inv(inv(P) + H'*inv(R)*H); % Mise a jour de la matrice de covariance
    
    %% Stockage du vecteur d'etat vrai et estime, et de la matrice  
    %% de covariance (pour l'affichage des resultats)
    xtrue_store = [xtrue_store, xtrue_new];
    xhat_store = [xhat_store, xhat];
    P_store = [P_store, P];
    Time_store = [Time_store, delta*k];

    xtrue = xtrue_new;
end

%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Affichages %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%

figure(1)
plot3(xtrue_store(1,1),xtrue_store(2,1),xtrue_store(3,1),'k*')
hold on
plot3(xtrue_store(1,:),xtrue_store(2,:),xtrue_store(3,:),'k-')
hold on
plot3(xhat_store(1,:),xhat_store(2,:),xhat_store(3,:),'b-')
hold on
plot3(xx,yy,zz,'r.');
grid on
xlabel('x [m]');
ylabel('y [m]');
zlabel('z [m]');
title('Trajectoire 3D du drone (noir, vrai; bleu estime) et balises (points rouges)')
axis equal

%% Determination des bandes 3 sigma a partir de la racine carree des elements 
%% sur la diagonale de P, afin de verifier la consistance de l'EKF. 
%% En theorie, l'erreur d'estimation devrait toujours rester a l'interieur de cette bande.

for k = 1:T
    Three_sigmaBd(:,k) = 3*sqrt(diag(P_store(:,9*(k-1)+1:9*k))); 
end

%%%%%% Position du drone %%%%%%

BandColor = [0.95,0.95,0.95];

figure(2)
subplot(3,1,1)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(1,:),fliplr(-Three_sigmaBd(1,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(1,:)-xhat_store(1,:),'b-')
ylabel('x [m]')
title('Erreur d''estimation de la position (bleu) et bande 3 sigma (zone grisee)')
xlim([0.101, T*delta-1e-3])
subplot(3,1,2)
fill([Time_store,fliplr(Time_store)], [Three_sigmaBd(2,:),fliplr(-Three_sigmaBd(2,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(2,:)-xhat_store(2,:),'b-')
ylabel('y [m]')
xlim([0.101, T*delta-1e-3])
subplot(3,1,3)
fill([Time_store,fliplr(Time_store)], [Three_sigmaBd(3,:),fliplr(-Three_sigmaBd(3,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(3,:)-xhat_store(3,:),'b-')
ylabel('z [m]')
xlim([0.101, T*delta-1e-3])
xlabel('temps [s]')

%%%%%% Vitesse du drone %%%%%%

figure(3)
subplot(3,1,1)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(4,:),fliplr(-Three_sigmaBd(4,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(4,:)-xhat_store(4,:),'b-')
ylabel('v_x')
xlim([0.101, T*delta-1e-3])
title('Erreur d''estimation de la vitesse (bleu) et bande 3 sigma (zone grisee)')
subplot(3,1,2)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(5,:),fliplr(-Three_sigmaBd(5,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(5,:)-xhat_store(5,:),'b-')
ylabel('v_y')
xlim([0.101, T*delta-1e-3])
subplot(3,1,3)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(6,:),fliplr(-Three_sigmaBd(6,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(6,:)-xhat_store(6,:),'b-')
xlim([0.101, T*delta-1e-3])
ylabel('v_z')
xlabel('temps [s]')

%%%%%% Acceleration du drone %%%%%%

figure(4)
subplot(3,1,1)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(7,:),fliplr(-Three_sigmaBd(7,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(7,:)-xhat_store(7,:),'b-')
xlim([0.101, T*delta-1e-3])
ylabel('a_x')
title('Erreur d''estimation de l''acceleration (bleu) et bande 3 sigma (zone grisee)')
subplot(3,1,2)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(8,:),fliplr(-Three_sigmaBd(8,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(8,:)-xhat_store(8,:),'b-')
xlim([0.101, T*delta-1e-3])
ylabel('a_y')
subplot(3,1,3)
fill([Time_store,fliplr(Time_store)],[Three_sigmaBd(9,:),fliplr(-Three_sigmaBd(9,:))],BandColor,'LineStyle','--');
hold on
plot(Time_store,xtrue_store(9,:)-xhat_store(9,:),'b-')
xlim([0.101, T*delta-1e-3])
ylabel('a_z')
xlabel('temps [s]')
