%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Mise en oeuvre du filtre de Kalman pour la fusion des mesures de vitesse
% angulaire d'un gyroscope et de position angulaire d'un compas magnetique 
% montes sur le meme plateau rotatif 
%
% - Gyroscope: capteur proprioceptif
% - Compas magnetique: capteur exteroceptif
% 
% Souce: Cours GLO-4001/GLO-7021, Philippe Giguere, Universite de Laval, Canada (2010-2012)
% www2.ift.ulaval.ca/~pgiguere/cours/IntroRobotique/index.html
% (adapte par F. Morbidi, Jan. 2019)
%
% TD2, Exercice III - Surveillance Distribuee de Systemes Multi-agents  
% --------------------------------------------------------------------------

clear
close all
clc

% Variables decrivant le systeme dynamique
Dk = 0.05;  % Pas de discretisation (plus les mesures sont frequentes, plus petite sera l'erreur) [s]
wrot = 0.2; % Frequence de rotation du plateau [rad/s]

Sigma_Gyro = 0.2;  % ecart type du bruit du gyroscope [deg/s]
Sigma_Compas = 10; % ecart type du bruit du compas [deg]
GyroBiais = 0.1;   % Biais du gyroscope (c'est-a-dire, mu_gyro)
xVrai = [0 0]';    % Initialisation de l'etat reel (position et vitesse angulaire du plateau)

A = [1 Dk; 0 0]; % Matrice dynamique 
B = [0; 1];      % Matrice d'entree
H = [1 0];       % Matrice de l'equation de mesure: nous avons "supprime" le gyroscope
Q = [0 0; 0 (4*Sigma_Gyro)^2];
R = Sigma_Compas^2; 

% Initialisation du filtre de Kalman
x = xVrai; % Etat estime initial
P = [Sigma_Compas^2 0; 0 Sigma_Gyro^2]; % Covariance de l'erreur d'estimation 

% Initialisation des mesures du gyro
xGyro = 0;

% Nombre d'iterations du filtre de Kalman
Tmax = 4000; 

for iStep = 1:Tmax
    % Simulation du systeme dynamique
    time = iStep*Dk;
  
    % On va faire tourner le plateau avec les deux capteurs
    xVrai = 40*[sin(wrot*time) wrot*cos(wrot*time)]';

    % On simule la reponse des capteurs
    compass = xVrai(1) + Sigma_Compas*randn;
    gyro = xVrai(2) + Sigma_Gyro*randn + GyroBiais;

    % Mesure et commande
    z = compass;
    u = gyro;
    
    % ==== Prediction du filtre de Kalman ====
    x = A*x + B*u;
    P = A*P*A' + Q;
    
    % ==== Correction du filtre de Kalman ====    
    K = P*H'/(H*P*H'+ R); % Gain de Kalman
    x = x + K*(z - H*x);
    P = (eye(2) - K*H)*P;
    
    % Integration des mesures du gyro, pour fin de comparaison
    xGyro = xGyro + gyro*Dk;
    xGyroInt(iStep) = xGyro;
    
    % Cueillette des donnees pour les graphiques
    Ax(iStep)           = xVrai(1);
    AEx(iStep)          = x(1);
    ACompas(iStep)      = z(1);
    AGyro(iStep)        = gyro;
    ErreurEx(iStep)     = x(1) - xVrai(1);
    ErreurCompas(iStep) = z(1) - xVrai(1);
    ATime(iStep) = time;
end

% On traca la distribution du bruit du compas (assume gaussien) et on fait
% la comparaison avec le resultat de l'application du filtre de Kalman

myFontSize = 12; % Pour changer la police dans les figures

figure(1)
subplot(2,1,1);
hist(ErreurCompas,20);
mean(ErreurCompas)
xlabel('Erreur compas [deg]','FontSize',myFontSize);
set(gca,'FontSize',myFontSize-2);
c = axis;

subplot(2,1,2);
hist(ErreurEx,20);
xlim(c(1:2));
mean(ErreurEx)
xlabel('Erreur filtre de Kalman [deg]','FontSize',myFontSize);
set(gca,'FontSize',myFontSize-2);

% Affichage des resultats: position reelle et estimee de la plateau 
% et mesures du gyro et du compas

figure(2)
subplot(3,1,1)
tl = 1; th = Tmax;
h(1) = plot(ATime(tl:th),Ax(tl:th),'k','LineWidth',2);
hold on
h(2) = plot(ATime(tl:th),ACompas(tl:th),'r*');
h(3) = plot(ATime(tl:th),xGyroInt(tl:th),'o','LineWidth',2);
h(4) = plot(ATime(tl:th),AEx(tl:th),'go','MarkerFaceColor','g'); 
plot(ATime(tl:th),AEx(tl:th),'go','MarkerFaceColor','g','MarkerSize',3)
plot(ATime(tl:th),Ax(tl:th),'k','LineWidth',2)
xlabel('temps [s]','FontSize',myFontSize)
ylabel('Angle [deg]','FontSize',myFontSize)
legend(h,{'Position reelle' 'Compas' 'Integrale Gyroscope' 'Filtre de Kalman'},'Location','NorthWest')
set(gca,'FontSize',myFontSize-2)

subplot(3,1,2)
tl = 1; th = 50;
h(1) = plot(ATime(tl:th),Ax(tl:th),'k','LineWidth',2);
hold on
h(2) = plot(ATime(tl:th),ACompas(tl:th),'r*');
h(3) = plot(ATime(tl:th),xGyroInt(tl:th),'o','LineWidth',2);
h(4) = plot(ATime(tl:th),AEx(tl:th),'go','MarkerFaceColor','g');
xlabel('temps [s]','FontSize',myFontSize);
ylabel('Angle [deg]','FontSize',myFontSize);
legend(h,{'Position reelle' 'Compas' 'Integrale Gyroscope' 'Filtre de Kalman'},'Location','NorthWest')
set(gca,'FontSize',myFontSize-2);

subplot(3,1,3)
tl = Tmax-50; th = Tmax;
h(1) = plot(ATime(tl:th),Ax(tl:th),'k','LineWidth',2);
hold on
h(2) = plot(ATime(tl:th),ACompas(tl:th),'r*');
h(3) = plot(ATime(tl:th),xGyroInt(tl:th),'o','LineWidth',2);
h(4) = plot(ATime(tl:th),AEx(tl:th),'go','MarkerFaceColor','g');
xlabel('temps [s]','FontSize',myFontSize)
ylabel('Angle [deg]','FontSize',myFontSize)
legend(h,{'Position reelle' 'Compas' 'Integrale Gyroscope' 'Filtre de Kalman'},'Location','NorthWest')
set(gca,'FontSize',myFontSize-2)
