%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Application de l'algorithme des moindres carres recursifs au probleme 
% d'ajustement de courbe. En particulier, l'objectif est d'estimer 
% les parameters (a, b) d'une droite d'equation y = a*t + b, à partir  
% de mesures bruitees y1,y2,...,yn, recues de maniere sequentielle 
%
% Pour plus de details, voir l'Exemple 3.6, pp. 92-93, dans le livre 
% "Optimal State Estimation: Kalman, H infinity, Nonlinear Approaches", 
% Dan Simon, Wiley, 2006

% TD2, Exercice I - Surveillance Distribuee de Systemes Multi-agents 
% -------------------------------------------------------------------------

clear
close all
clc

% Nombre de mesures dans le temps
n = 50; 
t = 0:n; 

% Parametres de la droite "ideale" (a estimer) 
b = 5; % Intersection de la droite avec l'axe vertical
a = 1; % Pente de la droite

% Generation de mesures entachees par le bruit
sigma = 1.5; % Ecart type du bruit blanc gaussien sur les mesures
R = sigma^2; % Variance du bruit blanc gaussien sur les mesures

for i=1:n
    y(i) = b + a*t(i) + sigma*randn; 
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%--- Initialisation de l'estimateur recursif ---
xhat = [0, 0]'; % Valeur initiale des parametres estimes
P = 0.5*eye(2); % Matrice de covariance initiale (confiance sur l'estime initial des parametres de la droite)
xhat_store = xhat;
P_store = P;

%%--- Mise a jour de l'estime xhat et de la covariance de l'erreur d'estimation P ---
for i=1:n
    H = [1, t(i)]; % Matrice de mesure
    K = P*H'/(H*P*H' + R); % Gain de l'estimateur (gain de Kalman)
    xhat = xhat + K*(y(i) - H*xhat); % Mise a jour de l'estime
    P = (eye(2) - K*H)*P; % Mise a jour de la matrice de covariance
    xhat_store = [xhat_store, xhat]; % Les vecteurs estimes sont stockes
    P_store = [P_store, P]; % Les matrices de covariance sont stockees
end

disp('Valeurs finales des parametres estimes de la droite:')
xhat_store(:,end)

%%%%%%%%%%%% Affichage des resultats %%%%%%%%%%%%%%%

figure(1)
subplot(2,1,1)
plot(0:n-1,xhat_store(1,1:n),'r-','LineWidth',1)
hold on
plot(0:n-1,b*ones(n,1),'b','LineWidth',1)
ylabel('b estime')
xlabel('temps')
title('Evolution de l''estimation de b et a (rouge) et vraies valeurs (bleu)')
subplot(2,1,2)
plot(0:n-1,xhat_store(2,1:n),'r-','LineWidth',1)
hold on
plot(0:n-1,a*ones(n,1),'b','LineWidth',1)
ylabel('a estime')
xlabel('temps')

%%%%%%%%%%%

figure(2)
plot(0:n-1,P_store(1,1:2:2*n-1),'k','LineWidth',1)
hold on
plot(0:n-1,P_store(2,2:2:2*n),'k--','LineWidth',1)
title('Evolution de la variance de l''erreur d''estimation (elements diagonaux de P)')
xlabel('temps')
ylabel('Variances')
lgd = legend('P(1,1)','P(2,2)');
lgd.FontSize = 12;

% ----------
% Remarque:
% ----------
% Les variances des erreurs d'estimation tendent asymptotiquement
% vers zero. Ca veut dire que nous avons de plus en plus de 
% confiance dans nos estimes, au fur et a mesure que l'on dispose 
% de plus de mesures 

%%%%%%%%%%%%

figure(3)
title('Evolution de la droite estimee (rouge) et mesures (bleu)')
xlabel('temps')
ylabel('y')
axis equal
axis([0 n 0 n])
for i=1:n
    hold on
    plot(t(i),y(i),'b*')
    plot(t,xhat_store(1,i) + xhat_store(2,i)*t,'r')
    pause(0.1)
end
